import numpy as np
import scipy as sp
import scipy.signal
from multiprocessing.pool import ThreadPool as Pool


def lfiltic_vec(b, a, y, x=None):
    """
    Construct initial conditions for lfilter given input and output vectors.

    Given a linear filter (b, a) and initial conditions on the output `y`
    and the input `x`, return the initial conditions on the state vector zi
    which is used by `lfilter` to generate the output given the input.

    Parameters
    ----------
    b : array_like
        Linear filter term.
    a : array_like
        Linear filter term.
    y : array_like
        Initial conditions.

        If ``N = len(a) - 1``, then ``y = {y[-1], y[-2], ..., y[-N]}``.

        If `y` is too short, it is padded with zeros.
    x : array_like, optional
        Initial conditions.

        If ``M = len(b) - 1``, then ``x = {x[-1], x[-2], ..., x[-M]}``.

        If `x` is not given, its initial conditions are assumed zero.

        If `x` is too short, it is padded with zeros.

    Returns
    -------
    zi : ndarray
        The state vector ``zi = {z_0[-1], z_1[-1], ..., z_K-1[-1]}``,
        where ``K = max(M, N)``.

    See Also
    --------
    lfilter, lfilter_zi

    """
    N = np.size(a) - 1
    M = np.size(b) - 1
    K = max(M, N)
    y = np.asarray(y)
    batch_size = y.shape[0]

    if y.dtype.kind in 'bui':
        # ensure calculations are floating point
        y = y.astype(np.float64)
    zi = np.zeros((batch_size, K), y.dtype)
    if x is None:
        x = np.zeros((batch_size, M), y.dtype)
    else:
        x = np.asarray(x)
        L = np.shape(x)[1]
        if L < M:
            x = np.r_[x, np.zeros((batch_size, M - L))]
    L = np.shape(y)[1]
    if L < N:
        y = np.r_[y, np.zeros((batch_size, N - L))]

    for m in range(M):
        zi[:, m] = np.sum(b[m + 1:] * x[:, :M - m], axis=1)

    for m in range(N):
        zi[:, m] -= np.sum(a[m + 1:] * y[:, :N - m], axis=1)

    return zi


def lfilter_mimo_channels_first(b, a, u_in):
    batch_size, in_ch, seq_len = u_in.shape
    out_ch, _, _ = a.shape
    y_out = np.zeros_like(u_in, shape=(batch_size, out_ch, seq_len))
    for out_idx in range(out_ch):
        for in_idx in range(in_ch):
            y_out[:, out_idx, :] += scipy.signal.lfilter(b[out_idx, in_idx, :], a[out_idx, in_idx, :],
                                                         u_in[:, in_idx, :], axis=-1)
    return y_out


def lfilter_mimo_components_channels_first(b, a, u_in):
    batch_size, in_ch, seq_len = u_in.shape
    out_ch, _, _ = a.shape
    y_comp_out = np.zeros_like(u_in, shape=(batch_size, out_ch, in_ch, seq_len))
    for out_idx in range(out_ch):
        for in_idx in range(in_ch):
            y_comp_out[:, out_idx, in_idx, :] = scipy.signal.lfilter(b[out_idx, in_idx, :], a[out_idx, in_idx, :], u_in[:, in_idx, :], axis=-1)
    return y_comp_out  # [B, O, I, T]


def lfilter_mimo(b, a, u_in):
    batch_size, seq_len, in_ch = u_in.shape
    out_ch, _, _ = a.shape
    y_out = np.zeros_like(u_in, shape=(batch_size, seq_len, out_ch))
    for out_idx in range(out_ch):
        for in_idx in range(in_ch):
            y_out[:, :, out_idx] += scipy.signal.lfilter(b[out_idx, in_idx, :], a[out_idx, in_idx, :],
                                                         u_in[:, :, in_idx], axis=-1)
    return y_out  # [B, T, O]


def lfilter_mimo_components(b, a, u_in):
    batch_size, seq_len, in_ch = u_in.shape
    out_ch, _, _ = a.shape
    y_comp_out = np.zeros_like(u_in, shape=(batch_size, seq_len, out_ch, in_ch))
    for out_idx in range(out_ch):
        for in_idx in range(in_ch):
            y_comp_out[:, :, out_idx, in_idx] = scipy.signal.lfilter(b[out_idx, in_idx, :], a[out_idx, in_idx, :], u_in[:, :, in_idx], axis=-1)
    return y_comp_out  # [B, T, O, I]


def lfilter_mimo_components_parallel(b, a, u_in):
    batch_size, seq_len, in_ch = u_in.shape
    out_ch, _, _ = a.shape
    T = u_in.shape[1]
    p = Pool(10)

    args_all = [(b[out_idx, in_idx, :], a[out_idx, in_idx], u_in[:, :, in_idx]) for out_idx in
                range(out_ch) for in_idx in range(in_ch)]

    def fun(args):
        return scipy.signal.lfilter(*args)
    res_lst = p.map(fun, args_all)
    y_comp_out = np.stack(res_lst, axis=-1).reshape(batch_size, T, out_ch, in_ch)

    return y_comp_out  # [B, T, O, I]


if __name__ == '__main__':

    batch_size = 5
    n_b = 2
    n_f = 2
    N = 100
    u_in = np.random.rand(batch_size, N, 1).astype(np.float)

    y_init = np.random.rand(*(batch_size, n_f)).astype(np.float)
    u_init = np.random.rand(*(batch_size, n_b)).astype(np.float)

    # coefficients of a 2nd order oscillator
    b_coeff = np.array([0.0706464146944544, 0], dtype=np.float)
    f_coeff = np.array([-1.87212998940304, 0.942776404097492], dtype=np.float)

    f_poly = np.r_[1.0, f_coeff]
    b_poly = np.r_[0.0, b_coeff]

    zi = lfiltic_vec(b_poly, f_poly, y_init, u_init)  # initial condition

    zi_loop = np.empty_like(zi)
    for batch_idx in range(batch_size):
        zi_loop[batch_idx, :] = sp.signal.lfiltic(b_poly, f_poly, y_init[batch_idx, :], u_init[batch_idx, :])  # initial condition


#    y_out = lfilter_mimo_components_channels_last
#    yout_vec, _ = sp.signal.lfilter(b_poly, f_poly, u_in, axis=0, zi=zi.T)

#    yout_loop = np.empty_like(yout_vec)
#    for batch_idx in range(batch_size):
#        yout_loop[:, batch_idx] = sp.signal.lfilter(b_poly, f_poly, u_in[:, batch_idx], axis=0, zi=zi[batch_idx, :])[0]




import torch
import numpy as np
import scipy as sp
import scipy.signal
from dynonet.filtering import lfilter_mimo_components


class MimoLinearDynamicalOperatorFun(torch.autograd.Function):
    r"""Applies a multi-input-multi-output linear dynamical filtering operation: :math:`y = G(u)`.

    Examples::

        >>> G = MimoLinearDynamicalOperatorFun.apply
        >>> n_b = 2
        >>> n_a = 2
        >>> N = 500
        >>> y_0 = torch.zeros(n_a, dtype=torch.double)
        >>> u_0 = torch.zeros(n_b, dtype=torch.double)
        >>> b_coeff = torch.tensor([0.0706464146944544, 0], dtype=torch.double, requires_grad=True)  # b_1, b_2
        >>> a_coeff = torch.tensor([-1.872112998940304, 0.942776404097492], dtype=torch.double, requires_grad=True)  # f_1, f_2
        >>> inputs = (b_coeff, a_coeff, u_in, y_0, u_0)
        >>> Y = G(*inputs)
        >>> print(Y.size())
        torch.Size([500, 1])
    """

    @staticmethod
    def forward(ctx, b_coeff, a_coeff, u_in, y_0=None, u_0=None):
        """
        In the forward pass we receive a Tensor containing the input and return
        a Tensor containing the output. ctx is a context object that can be used
        to stash information for backward computation. You can cache arbitrary
        objects for use in the backward pass using the ctx.save_for_backward method.
        """

        # detach tensors so we can cast to numpy

        b_coeff, a_coeff, u_in = b_coeff.detach(), a_coeff.detach(), u_in.detach()

        if y_0 is not None:
            y_0.detach()

        if u_0 is not None:
            u_0.detach()

        # useful parameters
        out_channels = b_coeff.shape[0]
        in_channels = b_coeff.shape[1]
        n_a = a_coeff.shape[2]

        # construct the A(q) polynomial with coefficient a_0=1
        a_poly = np.empty_like(a_coeff, shape=(out_channels, in_channels, n_a + 1))
        a_poly[:, :, 0] = 1
        a_poly[:, :, 1:] = a_coeff[:, :, :]
        b_poly = np.array(b_coeff)

        #zi = lfiltic_vec(b_poly, a_poly, y_init.numpy(), u_init.numpy())  # initial conditions for simulation

        y_out_comp = lfilter_mimo_components(b_poly, a_poly, u_in)  # [B, T, O, I]
        y_out = np.sum(y_out_comp, axis=-1)  # [B, T, O]
        y_out = torch.as_tensor(y_out, dtype=u_in.dtype)
        y_out_comp = torch.as_tensor(y_out_comp)

        ctx.save_for_backward(b_coeff, a_coeff, u_in, y_0, u_0, y_out_comp)
        return y_out

    @staticmethod
    def backward(ctx, grad_output):
        """
        In the backward pass we receive a Tensor containing the gradient of the loss
        with respect to the output, and we need to compute the gradient of the loss
        with respect to the input.
        """

        debug = False
        if debug:
            import pydevd  # required to debug the backward pass. Why?!?
            pydevd.settrace(suspend=False, trace_only_current_thread=True)

        b_coeff, a_coeff, u_in, y_0, u_0, y_out_comp = ctx.saved_tensors
        grad_b = grad_a = grad_u = grad_y0 = grad_u0 = None
        dtype_np = u_in.numpy().dtype

        out_channels, in_channels, n_b = b_coeff.shape
        _, _, n_a = a_coeff.shape
        batch_size, seq_len, _ = u_in.shape


        a_poly = np.empty_like(a_coeff, shape=(out_channels, in_channels, n_a + 1))
        a_poly[:, :, 0] = 1
        a_poly[:, :, 1:] = a_coeff[:, :, :]
        b_poly = np.array(b_coeff)  # not required?

        d0_np = np.array([1.0], dtype=dtype_np) #np.ones_like(u_in, shape=(out_channels, in_channels, 1))
        d1_np = np.array([0.0, 1.0], dtype=dtype_np)

        if ctx.needs_input_grad[0]:  # b_coeff
            # compute forward sensitivities w.r.t. the b_i parameters
            sens_b = np.zeros_like(u_in, shape=(batch_size, seq_len, out_channels, in_channels, n_b)) # [B, T, O, I, D]

            for out_idx in range(out_channels):  # it is like a lfilter_mimo_components, can be optimized
                for in_idx in range(in_channels):
                    sens_b[:, :, out_idx, in_idx, 0] = sp.signal.lfilter(d0_np, a_poly[out_idx, in_idx, :], u_in[:, :, in_idx])
            for idx_coeff in range(1, n_b):
                sens_b[:, idx_coeff:, :, :, idx_coeff] = sens_b[:, :-idx_coeff, :, :, 0]
            sens_b = torch.as_tensor(sens_b)
            grad_b = torch.einsum('bto,btoid->oid', grad_output, sens_b)  # einstein summation

        if ctx.needs_input_grad[1]:  # a_coeff
            # compute forward sensitivities w.r.t. the f_i parameters
            sens_a = np.zeros_like(u_in, shape=(batch_size, seq_len, out_channels, in_channels, n_a))
            for out_idx in range(out_channels): # it is like a lfilter_mimo_components, can be optimized
                for in_idx in range(in_channels):
                    sens_a[:, :, out_idx, in_idx, 0] = sp.signal.lfilter(d1_np, a_poly[out_idx, in_idx, :], -y_out_comp[:, :, out_idx, in_idx], axis=-1)

            for idx_coeff in range(1, n_a):
                sens_a[:, idx_coeff:, :, :, idx_coeff] = sens_a[:, :-idx_coeff, :, :, 0]
            sens_a = torch.as_tensor(sens_a)
            # compute vector-jacobian product for f
            grad_a =  torch.einsum('bto,btoid->oid', grad_output, sens_a)


        if ctx.needs_input_grad[2]: # u_in
            # compute jacobian w.r.t. u
            grad_output_flip = grad_output.numpy()[:, ::-1, :]  # [B, T, O]

            grad_u = np.zeros_like(u_in)  # [B, T, I]
            for in_idx in range(in_channels):
                for out_idx in range(out_channels):
                    grad_u[:, :, in_idx] += scipy.signal.lfilter(b_poly[out_idx, in_idx, :], a_poly[out_idx, in_idx, :], grad_output_flip[:, :, out_idx], axis=-1)
            grad_u = np.array(grad_u[:, ::-1, :]).astype(dtype_np)

            grad_u = torch.as_tensor(grad_u)

        return grad_b, grad_a, grad_u, grad_y0, grad_u0


if __name__ == '__main__':

    import numpy as np
    import matplotlib.pyplot as plt
    from torch.autograd import gradcheck
    from torch.autograd.gradcheck import get_numerical_jacobian, get_analytical_jacobian


    # copied from torch.autograd.gradcheck
    def istuple(obj):
        # Usually instances of PyStructSequence is also an instance of tuple
        # but in some py2 environment it is not, so we have to manually check
        # the name of the type to determine if it is a namedtupled returned
        # by a pytorch operator.
        t = type(obj)
        return isinstance(obj, tuple) or t.__module__ == 'torch.return_types'


    # copied from torch.autograd.gradcheck
    def _as_tuple(x):
        if istuple(x):
            return x
        elif isinstance(x, list):
            return tuple(x)
        else:
            return x,

    # In[Setup problem]
    in_ch = 1
    out_ch = 4
    n_b = 2
    n_a = 1
    batch_size = 8
    seq_len = 16

    # In[Create system]

    b_coeff = torch.tensor(np.random.randn(*(out_ch, in_ch, n_b)), requires_grad=True)
    a_coeff = torch.tensor(np.random.rand(*(out_ch, in_ch, n_a)), requires_grad=True)
    G = MimoLinearDynamicalOperatorFun.apply
    y_0 = torch.tensor(0*np.random.randn(*(out_ch, in_ch, n_a)))
    u_0 = torch.tensor(0*np.random.randn(*(out_ch, in_ch, n_b)))
    u_in = torch.tensor(1*np.random.randn(*(batch_size, seq_len, in_ch)), requires_grad=True)
    inputs = (b_coeff, a_coeff, u_in, y_0, u_0)

    # In[Forward pass]
    y_out = G(*inputs)

    # In[Finite difference derivatives computation]
    def G_fun(input):
        return _as_tuple(G(*input))[0]
    numerical = get_numerical_jacobian(G_fun, inputs)

    # In[Autodiff derivatives computation]
    analytical, reentrant, correct_grad_sizes = get_analytical_jacobian(inputs, y_out)
    #torch.max(numerical[0]- analytical[0])
    test = gradcheck(G, inputs, eps=1e-6, atol=1e-4, raise_exception=True)



import torch
import numpy as np  # temporary
from torch.nn.parameter import Parameter
from dynonet.functional import MimoLinearDynamicalOperatorFun


class MimoLinearDynamicalOperator(torch.nn.Module):
    r"""Applies a multi-input-multi-output linear dynamical filtering operation.

    Args:
        in_channels (int): Number of input channels
        out_channels (int): Number of output channels
        n_b (int): Number of learnable coefficients of the transfer function numerator
        n_a (int): Number of learnable coefficients of the transfer function denominator
        n_k (int, optional): Number of input delays in the numerator. Default: 0

    Shape:
        - Input: (batch_size, seq_len, in_channels)
        - Output: (batch_size, seq_len, out_channels)

    Attributes:
        b_coeff (Tensor): The learnable coefficients of the transfer function numerator
        a_coeff (Tensor): The learnable coefficients of the transfer function denominator

    Examples::

        >>> in_channels, out_channels = 2, 4
        >>> n_b, n_a, n_k = 2, 2, 1
        >>> G = MimoLinearDynamicalOperator(in_channels, out_channels, n_b, n_a, n_k)
        >>> batch_size, seq_len = 32, 100
        >>> u_in = torch.ones((batch_size, seq_len, in_channels))
        >>> y_out = G(u_in, y_0, u_0) # shape: (batch_size, seq_len, out_channels)
    """

    def __init__(self, in_channels, out_channels, n_b, n_a, n_k=0):
        super(MimoLinearDynamicalOperator, self).__init__()
        self.b_coeff = Parameter(torch.zeros(out_channels, in_channels, n_b))
        self.a_coeff = Parameter(torch.zeros(out_channels, in_channels, n_a))
        self.out_channels = out_channels
        self.in_channels = in_channels
        self.n_a = n_a
        self.n_b = n_b
        self.n_k = n_k

        with torch.no_grad():
            init_range = 0.01
            self.a_coeff[:] = (torch.rand(self.a_coeff.shape) - 0.5) * 2 * init_range
            self.b_coeff[:] = (torch.rand(self.b_coeff.shape) - 0.5) * 2 * init_range

    def forward(self, u_in, y_0=None, u_0=None):
        if self.n_k != 0:
            #u_d = u_in.roll(self.n_k, dims=-2)  # roll on the time axis
            #u_d[..., 0:self.n_k, :] = 0.0  # input sequence with delay
            u_d = torch.empty_like(u_in)
            u_d[..., self.n_k:, :] = u_in[:, :-self.n_k, :]
            u_d[..., 0:self.n_k, :] = 0.0
        else:
            u_d = u_in
        return MimoLinearDynamicalOperatorFun.apply(self.b_coeff, self.a_coeff, u_d, y_0, u_0)

    def get_filtdata(self):
        r"""Returns the numerator and denominator coefficients of the transfer function :math:`q^{-1}`-polynomials.

        The polynomials are function of the variable :math:`q^{-1}`.
        The polynomial coefficients b and a have length m and n, respectively and are sorted in descending power order.

        For a certain input channel :math:`i` and output channel :math:`o`, the  corresponding transfer
        function :math:`G_{i\rightarrow o}(z)` is:

        .. math::
            G_{i\rightarrow o}(z) = q^{-n_k}\frac{b[o, i, 0] + b[o, i, 1]q^{-1} + \dots + b[o, i, n]q^{-m+1}}
            {a[o, i, 0] + a[o, i, 1]q^{-1} + \dots + a[o, i, n]q^{-n+1}}

        Returns:
            np.array(in_channels, out_channels, m), np.array(in_channels, out_channels, n):
                numerator :math:`\beta` and denominator :math:`\alpha` polynomial coefficients of the transfer function.


        Examples::

            >>> num, den = G.get_tfdata()
            >>> G_tf = control.TransferFunction(G2_num, G2_den, ts=1.0)
        """
        return self.__get_filtdata__()

    def get_tfdata(self):
        r"""Returns the numerator and denominator coefficients of the transfer function :math:`z`-polynomials.

        The polynomials are function of the variable Z-transform variable :math:`z`.
        The polynomial coefficients :math::`\beta` and :math:`\alpha` have equal length p and are sorted in descending power order.

        For a certain input channel :math:`i` and output channel :math:`o`, the  corresponding transfer
        function :math:`G_{i\rightarrow o}(z)` is:

        .. math::
            G_{i\rightarrow o}(z) = \frac{\beta[o, i, 0]z^{n-1} + \beta[o, i, 1]z^{n-1} + \dots + \beta[o, i, p]}{\alpha[o, i, 0]z^{n-1} + \alpha[o, i, 1]z^{n-2} + \dots + \alpha[o, i, p]}

        Returns:
            np.array(in_channels, out_channels, p), np.array(in_channels, out_channels, p):
                numerator :math:`\beta` and denominator :math:`\alpha` polynomial coefficients of the transfer function.


        Examples::

            >>> num, den = G.get_tfdata()
            >>> G_tf = control.TransferFunction(G2_num, G2_den, ts=1.0)
        """
        return self.__get_tfdata__()

    def __get_filtdata__(self):
        # returns the coefficients of the polynomials b and a as function of q^{-1}
        b_coeff_np, a_coeff_np = self.__get_ba_coeff__()
        b_seq = np.zeros_like(b_coeff_np, shape=(self.out_channels, self.in_channels, self.n_b + self.n_k))  #b_coeff_np
        b_seq[:, :, self.n_k:] = b_coeff_np[:, :, :]
        a_seq = np.empty_like(a_coeff_np, shape=(self.out_channels, self.in_channels, self.n_a + 1))
        a_seq[:, :, 0] = 1
        a_seq[:, :, 1:] = a_coeff_np[:, :, :]
        return b_seq, a_seq

    def __get_tfdata__(self):
        b_seq, a_seq = self.__get_filtdata__()
        M = self.n_b + self.n_k  # number of numerator coefficients of the q^{-1} polynomial
        N = self.n_a + 1  # number of denominator coefficients of the q^{-1} polynomial
        if M > N:
            num = b_seq
            den = np.c_[a_seq, np.zeros((self.out_channels, self.in_channels, M - N))]
        elif N > M:
            num = np.c_[b_seq, np.zeros((self.out_channels, self.in_channels, N - M))]
            den = a_seq
        else:  # N == M
            num = b_seq
            den = a_seq

        return num, den

    def __get_ba_coeff__(self):
        return self.b_coeff.detach().numpy(), self.a_coeff.detach().numpy()


# SISO is implemented as a sub-case of MIMO
class SisoLinearDynamicalOperator(MimoLinearDynamicalOperator):
    r"""Applies a single-input-single-output linear dynamical filtering operation.

    Args:
        n_b (int): Number of learnable coefficients of the transfer function numerator
        n_a (int): Number of learnable coefficients of the transfer function denominator
        n_k (int, optional): Number of input delays in the numerator. Default: 0

    Shape:
        - Input: (batch_size, seq_len, 1)
        - Output: (batch_size, seq_len, 1)

    Attributes:
        b_coeff (Tensor): the learnable coefficients of the transfer function numerator
        a_coeff (Tensor): the learnable coefficients of the transfer function denominator

    Examples::

        >>> n_b, n_a = 2, 2
        >>> G = SisoLinearDynamicalOperator(b_coeff, a_coeff)
        >>> batch_size, seq_len = 32, 100
        >>> u_in = torch.ones((batch_size, seq_len))
        >>> y_out = G(u_in, y_0, u_0) # shape: (batch_size, seq_len, 1)
    """

    def __init__(self, n_b, n_a, n_k=0):
        super(SisoLinearDynamicalOperator, self).__init__(1, 1, n_b=n_b, n_a=n_a, n_k=n_k)

    def get_filtdata(self):
        b_seq, a_seq = super(SisoLinearDynamicalOperator, self).__get_filtdata__()  # MIMO numden
        return b_seq[0, 0, :], a_seq[0, 0, :]

    def get_tfdata(self):
        num, den = super(SisoLinearDynamicalOperator, self).__get_tfdata__()  # MIMO numden
        return num[0, 0, :], den[0, 0, :]


class MimoFirLinearDynamicalOperator(torch.nn.Module):
    r"""Applies a FIR linear multi-input-multi-output filtering operation.

    Args:
        in_channels (int): Number of input channels
        out_channels (int): Number of output channels
        n_b (int): Number of learnable FIR coefficients

    Shape:
        - Input: (batch_size, seq_len, in_channels)
        - Output: (batch_size, seq_len, out_channels)

    Attributes:
        G (torch.nn.Conv1d): The underlying Conv1D object used to implement the convolution

    Examples::

        >>> in_channels, out_channels = 2, 4
        >>> n_b = 128
        >>> G = MimoLinearDynamicalOperator(in_channels, out_channels, n_b)
        >>> batch_size, seq_len = 32, 100
        >>> u_in = torch.ones((batch_size, seq_len, in_channels))
        >>> y_out = G(u_in, y_0, u_0) # shape: (batch_size, seq_len, out_channels)
    """

    def __init__(self, in_channels, out_channels, n_b, channels_last=True):
        super(MimoFirLinearDynamicalOperator, self).__init__()
        self.G = torch.nn.Conv1d(in_channels, out_channels, kernel_size=n_b, bias=False, padding=n_b-1)
        #self.b_coeff = self.G.weight
        self.n_a = 0
        self.n_b = n_b
        self.out_channels = out_channels
        self.in_channels = in_channels
        self.channels_last = channels_last

    def forward(self, u_in):
        # PyTorch 1.4 does not have a channels_last option for Conv1d, which is more convenient
        # in our block-oriented modeling framework.
        # Then, let us transpose the last and second last channels manually before and after applying torch.nn.Conv1d
        if self.channels_last:
            u_torch = u_in.transpose(-2, -1)

        y_out = self.G(u_torch)
        y_out = y_out[..., 0:-self.n_b+1]

        if self.channels_last:
            y_out = y_out.transpose(-2, -1)
        return y_out

    def get_filtdata(self):
        return self.__get_filtdata__()

    def get_tfdata(self):
        return self.__get_tfdata__()

    def __get_filtdata__(self):
        b_coeff, a_coeff = self.__get_ba_coeff__()
        b_seq = b_coeff
        a_seq = np.empty_like(a_coeff, shape=(self.out_channels, self.in_channels, self.n_a + 1))
        a_seq[:, :, 0] = 1
        a_seq[:, :, 1:] = a_coeff[:, :, :]
        return b_seq, a_seq

    def __get_tfdata__(self):
        b_seq, a_seq = self.__get_filtdata__()
        M = self.n_b  # numerator coefficients
        N = self.n_a + 1  # denominator coefficients
        if M > N:
            num = b_seq
            den = np.c_[a_seq, np.zeros((self.out_channels, self.in_channels, M - N))]
        elif N > M:
            num = np.c_[self.b_poly, np.zeros((self.out_channels, self.in_channels, N - M))]
            den = a_seq
        else:  # N == M
            num = b_seq
            den = a_seq

        return num, den

    def __get_ba_coeff__(self):
        b_coeff_np = self.G.weight.detach().numpy()
        b_coeff_np = b_coeff_np[:, :, ::-1]
        a_coeff_np = np.zeros_like(b_coeff_np, shape=(self.out_channels, self.in_channels, 0))
        return b_coeff_np, a_coeff_np


class SisoFirLinearDynamicalOperator(MimoFirLinearDynamicalOperator):
    r"""Applies a FIR linear single-input-single-output filtering operation.

    Args:
        n_b (int): Number of learnable FIR coefficients

    Shape:
        - Input: (batch_size, seq_len, 1)
        - Output: (batch_size, seq_len, 1)

    Attributes:
        G (torch.nn.Conv1d): The underlying Conv1D object used to implement the convolution

    Examples::

        >>> n_b = 128
        >>> G = SisoFirLinearDynamicalOperator(n_b)
        >>> batch_size, seq_len = 32, 100
        >>> u_in = torch.ones((batch_size, seq_len, 1))
        >>> y_out = G(u_in, y_0, u_0) # shape: (batch_size, seq_len, 1)
    """
    def __init__(self, n_b, channels_last=True):
        super(SisoFirLinearDynamicalOperator, self).__init__(1, 1, n_b, channels_last=channels_last)

    def get_filtdata(self):
        b_seq, a_seq = super(SisoFirLinearDynamicalOperator, self).__get_filtdata__() # call to MIMO ba
        return b_seq[0, 0, :], a_seq[0, 0, :]

    def get_tfdata(self):
        num, den = super(SisoFirLinearDynamicalOperator, self).__get_tfdata__() # call to MIMO numden
        return num[0, 0, :], den[0, 0, :]


class StableSecondOrderMimoLinearDynamicalOperator(torch.nn.Module):
    r"""Applies a stable second-order linear multi-input-multi-output filtering operation.
    The denominator of the transfer function is parametrized in terms of two complex conjugate poles with magnitude
    :math:: `r, 0 < r < 1` and phase :math:: `\beta, < 0 \beta < \pi`. In turn, :math:: `r` and :math:: `\beta` are
    parametrized in terms of unconstrained variables :math:: `\rho` and :math:: `\psi`

    Args:
        in_channels (int): Number of input channels
        out_channels (int): Number of output channels

    Shape:
        - Input: (batch_size, seq_len, 1)
        - Output: (batch_size, seq_len, 1)

    Attributes:
        rho (Tensor): the learnable :math:: `\rho` coefficients of the transfer function denominator
        psi (Tensor): the learnable :math:: `\psi` coefficients of the transfer function denominator
        b_coeff (Tensor): the learnable numerator coefficients

    Examples::

        >>> in_channels = 2
        >>> out_channels = 4
        >>> G = StableSecondOrderMimoLinearDynamicalOperator(in_channels, out_channels)
        >>> batch_size, seq_len = 32, 100
        >>> u_in = torch.ones((batch_size, seq_len, in_channels))
        >>> y_out = G(u_in, y_0, u_0) # shape: (batch_size, seq_len, 1)

    """

    def __init__(self, in_channels, out_channels):
        super(StableSecondOrderMimoLinearDynamicalOperator, self).__init__()
        self.b_coeff = Parameter(torch.zeros(out_channels, in_channels, 2))
        self.rho = Parameter(torch.zeros(out_channels, in_channels, 1))
        self.psi = Parameter(torch.zeros((out_channels, in_channels, 1)))
        with torch.no_grad():
            self.rho[:] = torch.randn(self.rho.shape) * 0.1
            self.psi[:] = torch.randn(self.rho.shape) * 0.1
            self.b_coeff[:] = torch.randn(self.b_coeff.shape) * 0.01

    def forward(self, u_in, y_0=None, u_0=None):
        r = torch.sigmoid(self.rho)
        beta = np.pi * torch.sigmoid(self.psi)
        a_1 = -2 * r * torch.cos(beta)
        a_2 = r ** 2
        a_coeff = torch.cat((a_1, a_2), dim=-1)
        return MimoLinearDynamicalOperatorFun.apply(self.b_coeff, a_coeff, u_in, y_0, u_0)


class StableSecondOrderSisoLinearDynamicalOperator(StableSecondOrderMimoLinearDynamicalOperator):
    r"""Applies a stable second-order linear single-input-single-output filtering operation.
    The denominator of the transfer function is parametrized in terms of two complex conjugate poles with magnitude
    :math:: `r, 0 < r < 1` and phase :math:: `\beta, < 0 \beta < \pi`. In turn, :math:: `r` and :math:: `\beta` are
    parametrized in terms of unconstrained variables :math:: `\rho` and :math:: `\psi`

    """

    def __init__(self):
        super(StableSecondOrderSisoLinearDynamicalOperator, self).__init__(1, 1)  # in_channels, out_channels, n_b, n_a



import numpy as np
#np.array


def r_squared(y_true, y_pred, time_axis=0):
    """ Computes the R-square index.

    The R-squared index is computed separately on each channel.

    Parameters
    ----------
    y_true : np.array
        Array of true values.  If must be at least 2D.
    y_pred : np.array
        Array of predicted values.  If must be compatible with y_true'
    time_axis : int
        Time axis. All other axes define separate channels.

    Returns
    -------
    r_squared_val : np.array
        Array of r_squared value.
    """

    SSE = np.sum((y_pred - y_true)**2, axis=time_axis)
    y_mean = np.mean(y_true, axis=time_axis, keepdims=True)
    SST = np.sum((y_true - y_mean)**2, axis=time_axis)

    return 1.0 - SSE/SST


def error_rmse(y_true, y_pred, time_axis=0):
    """ Computes the Root Mean Square Error (RMSE).

    The RMSE index is computed separately on each channel.

    Parameters
    ----------
    y_true : np.array
        Array of true values.  If must be at least 2D.
    y_pred : np.array
        Array of predicted values.  If must be compatible with y_true'
    time_axis : int
        Time axis. All other axes define separate channels.

    Returns
    -------
    RMSE : np.array
        Array of r_squared value.

    """

    SSE = np.mean((y_pred - y_true)**2, axis=time_axis)
    RMSE = np.sqrt(SSE)
    return RMSE


def error_mean(y_true, y_pred, time_axis=0):
    """ Computes the error mean value.

    The RMSE index is computed separately on each channel.

    Parameters
    ----------
    y_true : np.array
        Array of true values.  If must be at least 2D.
    y_pred : np.array
        Array of predicted values.  If must be compatible with y_true'
    time_axis : int
        Time axis. All other axes define separate channels.

    Returns
    -------
    e_mean : np.array
        Array of error means.
    """

    e_mean = np.mean(y_true - y_pred, axis=time_axis)
    return e_mean


def error_mae(y_true, y_pred, time_axis=0):
    """ Computes the error Mean Absolute Value (MAE)

    The RMSE index is computed separately on each channel.

    Parameters
    ----------
    y_true : np.array
        Array of true values.  If must be at least 2D.
    y_pred : np.array
        Array of predicted values.  If must be compatible with y_true'
    time_axis : int
        Time axis. All other axes define separate channels.

    Returns
    -------
    e_mean : np.array
        Array of error mean absolute values.
    """

    e_mean = np.mean(np.abs(y_true - y_pred), axis=time_axis)
    return e_mean

def fit_index(y_true, y_pred, time_axis=0):
    """ Computes the per-channel fit index.

    The fit index is commonly used in System Identification. See the definitionin the System Identification Toolbox
    or in the paper 'Nonlinear System Identification: A User-Oriented Road Map',
    https://arxiv.org/abs/1902.00683, page 31.
    The fit index is computed separately on each channel.

    Parameters
    ----------
    y_true : np.array
        Array of true values.  If must be at least 2D.
    y_pred : np.array
        Array of predicted values.  If must be compatible with y_true'
    time_axis : int
        Time axis. All other axes define separate channels.

    Returns
    -------
    fit_val : np.array
        Array of r_squared value.

    """

    err_norm = np.linalg.norm(y_true - y_pred, axis=time_axis, ord=2)  # || y - y_pred ||
    y_mean = np.mean(y_true, axis=time_axis)
    err_mean_norm = np.linalg.norm(y_true - y_mean, ord=2)  # || y - y_mean ||
    fit_val = 100*(1 - err_norm/err_mean_norm)

    return fit_val


if __name__ == '__main__':
    N = 20
    ny = 2
    SNR = 10
    y_true = SNR*np.random.randn(N, 2)
    y_pred = np.copy(y_true) + np.random.randn(N, 2)
    err_rmse_val = error_rmse(y_pred, y_true)
    r_squared_val = r_squared(y_true, y_pred)
    fit_val = fit_index(y_true, y_pred)

    print(f"RMSE: {err_rmse_val}")
    print(f"R-squared: {r_squared_val}")
    print(f"fit index: {fit_val}")



import torch
import torch.nn as nn


class MimoStaticNonLinearity(nn.Module):
    r"""Applies a Static MIMO non-linearity.
    The non-linearity is implemented as a feed-forward neural network.

    Args:
        in_channels (int): Number of input channels
        out_channels (int): Number of output channels
        n_hidden (int, optional): Number of nodes in the hidden layer. Default: 20
        activation (str): Activation function. Either 'tanh', 'relu', or 'sigmoid'. Default: 'tanh'

    Shape:
        - Input: (..., in_channels)
        - Output: (..., out_channels)

    Examples::

        >>> in_channels, out_channels = 2, 4
        >>> F = MimoStaticNonLinearity(in_channels, out_channels)
        >>> batch_size, seq_len = 32, 100
        >>> u_in = torch.ones((batch_size, seq_len, in_channels))
        >>> y_out = F(u_in, y_0, u_0) # shape: (batch_size, seq_len, out_channels)
    """

    def __init__(self, in_channels, out_channels, n_hidden=20, activation='tanh'):
        super(MimoStaticNonLinearity, self).__init__()

        activation_dict = {'tanh': nn.Tanh, 'relu': nn.ReLU, 'sigmoid': nn.Sigmoid}

        self.net = nn.Sequential(
            nn.Linear(in_channels, n_hidden),
            activation_dict[activation](),  #nn.Tanh(),
            nn.Linear(n_hidden, out_channels)
        )

    def forward(self, u_lin):
        y_nl = self.net(u_lin)
        return y_nl


class SisoStaticNonLinearity(MimoStaticNonLinearity):
    r"""Applies a Static SISO non-linearity.
    The non-linearity is implemented as a feed-forward neural network.

    Args:
        n_hidden (int, optional): Number of nodes in the hidden layer. Default: 20
        activation (str): Activation function. Either 'tanh', 'relu', or 'sigmoid'. Default: 'tanh'
        s
    Shape:
        - Input: (..., in_channels)
        - Output: (..., out_channels)

    Examples::

        >>> F = SisoStaticNonLinearity(n_hidden=20)
        >>> batch_size, seq_len = 32, 100
        >>> u_in = torch.ones((batch_size, seq_len, in_channels))
        >>> y_out = F(u_in, y_0, u_0) # shape: (batch_size, seq_len, out_channels)
    """
    def __init__(self, n_hidden=20, activation='tanh'):
        super(SisoStaticNonLinearity, self).__init__(in_channels=1, out_channels=1, n_hidden=n_hidden, activation=activation)


class MimoChannelWiseNonLinearity(nn.Module):
    r"""Applies a Channel-wise non-linearity.
    The non-linearity is implemented as a set of feed-forward neural networks (each one operating on a different channel).

    Args:
        channels (int): Number of both input and output channels
        n_hidden (int, optional): Number of nodes in the hidden layer of each network. Default: 10

    Shape:
        - Input: (..., channels)
        - Output: (..., channels)

    Examples::

        >>> channels = 4
        >>> F = MimoChannelWiseNonLinearity(channels)
        >>> batch_size, seq_len = 32, 100
        >>> u_in = torch.ones((batch_size, seq_len, channels))
        >>> y_out = F(u_in, y_0, u_0) # shape: (batch_size, seq_len, channels)

    """

    def __init__(self, channels, n_hidden=10):
        super(MimoChannelWiseNonLinearity, self).__init__()

        self.net = nn.ModuleList()
        for channel_idx in range(channels):
            channel_net = nn.Sequential(
                nn.Linear(1, n_hidden),  # 2 states, 1 input
                nn.ReLU(),
                nn.Linear(n_hidden, 1)
            )
            self.net.append(channel_net)

    def forward(self, u_lin):

        y_nl = []
        for channel_idx, u_channel in enumerate(u_lin.split(1, dim=-1)):  # split over the last dimension (input channel)
            y_nl_channel = self.net[channel_idx](u_channel)  # Process blocks individually
            y_nl.append(y_nl_channel)

        y_nl = torch.cat(y_nl, -1)  # concatenate all output channels
        return y_nl


if __name__ == '__main__':

    channels = 4
    nn1 = MimoChannelWiseNonLinearity(channels)
    in_data = torch.randn(100, 10, channels)
    xx = net_out = nn1(in_data)